/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                               fplot.h
 *
 *  Defines all plot objects, including mute (file dump) plots, xmgr/xmgrace
 *  command pipes and gl/Ygl X-windows graphs
 *
 *                           Class Hierarchy
 *                           ^^^^^^^^^^^^^^^
 *           /-- StateDump                          /----  MutePlot2D
 *           |                                      |
 *           |-- MutePointPlot    /-- FieldPlot2D --|----  Plot2D
 *           |                    |                 
 *  PlotObj -|-- FieldPlotObj ----|-- FieldDump        /----  RasterPlot
 *           |                    |                    |
 *           |-- PointPlot        \-- FieldPlot1D \----|----  Plot1D
 *            \                                    |   |
 *             \_____ XmgrPointPlot                |   \----  MutePlot1D
 *             /                                   | 
 *            /                                    \______ 
 *  XmgrPlot --------------------------------------------- Xmgr1Dplot
 *
 *
 ******************************************************************************/

#ifndef CALC_FPLOT_H_included
#define CALC_FPLOT_H_included

#define METHOD_YGL  0
#define METHOD_XMGR 1
#define METHOD_MUTE 2

#define XMGR_DIVISIONS 6

#define xmarg_color_panel 50
#define ymarg_color_panel 15
#define dy_color_panel 10
#define color_num 150

#define pref_width 400
#define pref_height 200
#define xmargin 15
#define ymargin 15
#define char_height 8
#define char_width  6

#define bgr_color 0
#define plot_color 1
#define grid_color 2
#define text_color 3
#define label_color 4

#define pixelspergridline 45
#define line_color 5

#define dxmargin 0.14
#define dymargin 0.21

#define dx_legend 0.2
#define dy_legend 0.1

#define title_color 4  // landscape mode

#ifdef _YGL_INCLUDED_
void   mouse_pause();
#endif

double format_double(double number0, double number1, int divisions, int sigdigits, char *fstr);
double format_digits(double number, int &digits);

class SimulationObj; // forward class declaration 

//*******************************************************************************
//                 C L A S S   P L O T O B J E C T
//*******************************************************************************

class PlotObj
{
protected:
public:

  static int UPDATE_STEPS;
  static int UPDATE_STEPS_1D;
  static double UPDATE_ACCURACY;

  int   plot_height, win_height, plot_width,  win_width;
  char  win_title[255];
  long  winid;

  char     x_label[80];
  double   x_value;

  char  log_plot, mute;

  double   fmin, fmax;
  double   *fptr;

  PlotObj(double *ptr, char islog, const char *txt = "", const char *label = "")
    {
    fptr = ptr; log_plot = islog;
    if (txt) strcpy(win_title,txt); else win_title[0] = 0;
    if (label) strcpy(x_label,label); else x_label[0] = 0;
    x_value = fmax = fmin = 0.0;
    };

  virtual ~PlotObj() { };

  virtual void draw() = 0;
  virtual void redraw() = 0;
  void    draw_grid(double, double, double, double, char);

  virtual double get_value(long ind = 0)
   {
   double v = fptr[ind];
   if (!log_plot) return v;
      else return ( (v <= 0.0) ? 0.0 : log(v)/log(10.0) );
   };

};

//*******************************************************************************
//                   C L A S S    S T A T E    D U M P
//*******************************************************************************

class StateDump : public PlotObj
{
protected:

  bool   complete;
  double exportTime;
  char   fileName[255];
  class  SimulationObj *Sim;

public:

 FILE   *file;

 StateDump( SimulationObj *sim, double t,
            const char *fname = 0, const char *txt = 0) : PlotObj(0, 0, txt) {
    strcpy(fileName, fname); 
    Sim = sim;
    exportTime = t;
    complete = false;
    };

 ~StateDump() { }

 void    draw();
 void    redraw() { draw(); }
};
//*******************************************************************************
//                  C L A S S   P O I N T   P L O T
//*******************************************************************************

class PointPlot : public PlotObj
{
protected:
  
  long   bufferSteps;
  double *buffer;
  double *times;
  double *timeptr;
  double last;
  int    count;

public:

 PointPlot(double *ptr, double *tptr, char islog, double t, 
          const char *txt = 0, const char *label = "time=", 
          int hor = pref_width, int ver = pref_height);

 ~PointPlot() { delete [] buffer; delete [] times; };

 void    draw();
 void    redraw()  { if (count) count --; draw(); }

 void    set_colors();
 void    get_range();

};

//*******************************************************************************
//               C L A S S   M U T E   P O I N T   P L O T
//*******************************************************************************

class MutePointPlot : public PlotObj
{
protected:

  double *time;
  double fold, x_temp, f_temp;
  double tscale;
  char   fileName[255];
  double tHold, fHold;

public:

 FILE   *file;

 MutePointPlot(double *ptr, double *tptr, char islog,  double t, 
               const char *fname = "", const char *txt = "");

 ~MutePointPlot() { fclose(file); };

 void    draw();
 void    redraw()  { x_value = 0; draw(); }
};


//*******************************************************************************
//                 C L A S S   F I E L D P L O T O B J E C T
//*******************************************************************************

class FieldPlotObj : public PlotObj
{
protected:
public:

 FieldObj *field; 
 long     field_index;

 FieldPlotObj(FieldObj *f, char islog, const char *txt = 0, const char *label = 0):
   PlotObj((*f)(), islog, txt, label)  { field = f; field_index = 0; };

 double get_value(long ind = 0);

};
//*******************************************************************************

class FieldDump : public FieldPlotObj
{
protected:

  bool   complete;
  double exportTime;
  char   fileName[255];

public:

 FILE   *file;

 FieldDump(FieldObj *f, double t, const char *fname = 0, const char *txt = 0) : FieldPlotObj(f, 0, txt)
    {
    strcpy(fileName, fname);
    exportTime = t;
    complete = false;
    }

 // ~FieldDump() { }

 void    draw()  {
   if (field->Time < exportTime) { complete = false; return; }
   if ( complete ) return;
   complete = true;
   field->Export(fileName);
   };

 void    redraw() { draw(); }
};

//*******************************************************************************
//                      C L A S S   1 D - P L O T
//*******************************************************************************

class FieldPlot1D : public FieldPlotObj
{
protected:

  long    incr, num;
  double  *grid, *coord;

public:

 FieldPlot1D(FieldObj *f, char islog, const char *dir, double coord1, double coord2, 
             const char *label = "time=");

 virtual void    get_range();
};


//*******************************************************************************

class MutePlot1D : public FieldPlot1D
{
public:

 char   fileName[255];
 double total;
 double tscale;
 double exportTime;
 bool   complete;

 MutePlot1D(FieldObj *f, char islog, const char *dir, double coord1, double coord2, double T,
            const char *fname = 0, const char *txt = 0) :
 FieldPlot1D(f, islog, dir, coord1, coord2)  {
   exportTime = total = T;
   x_value = -1e-12;
   complete = false;
   tscale = double(UPDATE_STEPS_1D) / total;
   strcpy(fileName,fname);
   };

 ~MutePlot1D() { };

 void    draw();
 void    redraw() { x_value = 0.0; draw(); }; 

};

//*******************************************************************************

class Plot1D : public FieldPlot1D
{
protected:

  double tscale;

public:

 Plot1D(FieldObj *f, char islog, const char *dir, double coord1, double coord2,
        double T, int hor = pref_width, int ver = pref_height, 
        const char *label = "time=");

 void    draw();
 void    redraw() { x_value = 0.0; draw(); }; 
 void    set_colors();

};

//*******************************************************************************
//                  C L A S S   R A S T E R   P L O T
//*******************************************************************************

class RasterPlot : public FieldPlot1D
{
protected:

  double *raster;
  double *times;
  double last;
  int    count;

  int color1;  

  char    fixed;
  double  max_fixed, min_fixed;
  double  df_dcol, df_dcol_fixed;

public:

 RasterPlot(FieldObj *f, char islog, const char *dir, double coord1, double coord2, double t, 
  int hor = pref_width, int ver = pref_height, double min = 0.0, double max = -1.0,
  const char *label = "time=");

 ~RasterPlot() {  delete raster; delete times; };

 void    draw();
 void    redraw();
 void    set_colors();
 void    get_range();
 int     get_color(double);
 void    color_panel();
};


//*******************************************************************************
//                C L A S S   2 D   F I E L D   P L O T
//*******************************************************************************

class FieldPlot2D : public FieldPlotObj
{
protected:

  long    incr1, incr2, num1, num2;
  double  *grid1, *grid2;
  double  *coord1, *coord2;

public:

 FieldPlot2D(FieldObj *f, char islog, const char *dir, double coord, const char *label = "time=");

 virtual void  get_range();
};

//*******************************************************************************

class MutePlot2D : public FieldPlot2D
{
public:

 bool   complete;
 double exportTime;
 char   fileName[255];

 MutePlot2D(FieldObj *f, char islog, const char *dir, double coord, double T,
            const char *fname = 0, const char *txt = 0) :
   FieldPlot2D(f, islog, dir, coord)
   {
   exportTime = T;
   strcpy(fileName,fname);
   };

 ~MutePlot2D() {  };

 void    draw() {  if (field->Time < exportTime) { complete = false; return; }
                   if ( complete ) return;
                   complete = true;
                   long ind = field_index;
                   FILE *file = fopenAssure(fileName, "w", "plot", win_title);
		   for (int j = 0; j < num2; j++) {
		     for (int i = 0; i < num1; i++) {
                        ind = field_index + j * incr2 + i * incr1;
                        fprintf(file,"%.6g %.6g %.9g \n", coord1[i], coord2[j], get_value(ind));
		     }
		     fprintf(file,"\n");
		   }
                   fclose(file);
 }

 void    redraw() { draw(); }; 
};
//*******************************************************************************

class Plot2D : public FieldPlot2D
{
protected:

  int     color1;  
  char    fixed;
  double  max_fixed, min_fixed;
  double  df_dcol, df_dcol_fixed;

  double tscale;

public:

 Plot2D(FieldObj *f, char islog, const char *dir, double coord, double,  
        int hor = pref_width, int ver = pref_height, double min = 0.0, double max = -1.0,
        const char *label = "time=");

 void    draw();
 void    redraw();
 void    set_colors();
 int     get_color(double);
 void    color_panel();

 void    get_range() {  FieldPlot2D::get_range();
                             df_dcol = (fmax - fmin) / double(color_num);
                          }
};


//*******************************************************************************
//           B A S E   C L A S S   X M G R   P I P E   P L O T
//*******************************************************************************
// xmgr colors: 0-white 1-black 2-red 3-green 4-blue 5-yellow 6-brown 7-gray 
// 8-violet 9-cyan 10-magenta 11-orange 12-indigo 13-maroon 14-turquise 15 -green4

class XmgrPlot
{
protected:

  static int XMGR_STEPS;

  static int graph_num;
  static int graph_count;

  static int rows;
  static int cols;

  static FILE *file;

  int    steps;

  int    set_num;
  int    set_count;
  int    graph_id;

  int    view_left, view_right, view_top, view_bottom;

  double first, last;

  bool   logScale;

public:

 static void init(TokenString &, int gnum, FILE *f = stdout, int = 0, int = 0);
 static void processStrings(TokenString &);

 XmgrPlot(int sets, char is_log, const char *xlabel = "time");
 ~XmgrPlot() {};

 int setColor(int);
};

//*******************************************************************************
//               C L A S S   X M G R   P I P E   P L O T
//*******************************************************************************
// xmgr colors: 0-white 1-black 2-red 3-green 4-blue 5-yellow 6-brown 7-gray 
// 8-violet 9-cyan 10-magenta 11-orange 12-indigo 13-maroon 14-turquise 15 -green4

class XmgrPointPlot : public XmgrPlot, public PlotObj
{
protected:

  double **tptrs, **fptrs;
  double *x_old;
  double *f_old;
  double *x_temp;
  double *f_temp;

  double tHold, fHold;

  double **tArray, **fArray;
  int    *pointCount, *maxPoints;

  double tScale;

public:

 XmgrPointPlot(int sets, char islog, double T = 0, const char *xlabel = "time");

 void set_set(double *p, double *t, const char *txt = 0);

 ~XmgrPointPlot() { for (int i=0; i<set_num; i++) { delete [] tArray[i]; delete [] fArray[i]; }
                    delete [] tArray; delete [] fArray; delete [] pointCount; delete [] maxPoints;
                    delete [] tptrs; delete [] fptrs; delete [] x_old; delete [] f_old; delete [] x_temp; delete [] f_temp; 
                   } 

 void    get_range();
 void    draw();
 void    redraw()  { for (int i = 0; i < set_num; i++) x_old[i] = x_temp[i] = 0.0; draw(); }
 void    truncate(int set);
 void    addMemory(int);
};


//*******************************************************************************
//            C L A S S   1 D   X M G R   P I P E   P L O T  
//*******************************************************************************

class Xmgr1Dplot : public XmgrPlot, public FieldPlot1D
{
protected:

 double setTime[30];

public:

 Xmgr1Dplot(FieldObj *f, char islog, const char *dir, double coord1, double coord2, double T, 
            int m = 5);

 void    draw();
 void    redraw() { /*x_value = 0.0;*/ draw(); }; 
 int     setColor(int);
};

//*******************************************************************************
//                      C L A S S   P L O T   A R R A Y
//*******************************************************************************

class PlotArray
{
 protected:

 public:

  PlotObj **array;
  int     plot_num;
  int     count;
  char    ygl_on, method;

  PlotArray(class SimulationObj  &);

  int get_plot_num(TokenString &);

  PlotArray(int n)   { count = 0; ygl_on = 0;
                       plot_num = n;
                       if (n) array = new PlotObj *[n]; }

  ~PlotArray();

  void set_plot(PlotObj *plot)  {
     if (count >= plot_num) 
        throw makeMessage("in PlotArray: cannot set plot #%d: total number = %d", count, plot_num); 
     else array[count++] = plot;
     }

  void draw_plot(int n)   { if (n < count) array[n]->draw();  }
  void redraw_plot(int n)  { if (n < count) array[n]->redraw(); }
  void draw_all()    {  if (count) for (int i=0; i<count; i++) array[i]->draw(); }
  void redraw_all()  {  if (count) for (int i=0; i<count; i++) array[i]->redraw();  }

  void tail_draw()   { 
#ifdef	_YGL_INCLUDED_
  if (VERBOSE && ygl_on) 
    fprintf(stdlog, "\n==== Click the right mouse button to exit, left button to refresh plots ====\n");

    while(ygl_on)  {
      while(! (getbutton(RIGHTMOUSE) | getbutton(LEFTMOUSE) | getbutton(MIDDLEMOUSE) ) );
      if ( getbutton(RIGHTMOUSE) ) return;
      redraw_all(); 
      while( getbutton(RIGHTMOUSE) | getbutton(LEFTMOUSE) | getbutton(MIDDLEMOUSE) );
    }
#endif
    }
};

//*****************************************************************************
//                      S T A T U S      W I D G E T
//*****************************************************************************


class RunStatusString {

  int    statusLength, widgetLength, widgetCount;
  char   *widgetStr, *wholeStr;
  char   *timeStr, *extraStr;
  double *timePtr, *extraPtr;
  double total, t0;

 public:

  RunStatusString(int length, const char *s1, double *p1, double T, char *s2=0, double *p2=0);
  ~RunStatusString();
  
  void reset(double T = 0);
  void update(char c = 0);
};



//*******************************************************************************

#endif
